@@ -1,6 +1,6 @@
 /* wait-private.c 
    Copyright (C) 2000 Werner Koch (dd9jn)
-   Copyright (C) 2001, 2002, 2003, 2004 g10 Code GmbH
+   Copyright (C) 2001, 2002, 2003, 2004, 2005 g10 Code GmbH
  
    This file is part of GPGME.
  
@@ -29,7 +29,7 @@
 #include "context.h"
 #include "wait.h"
 #include "ops.h"
-#include "io.h"
+#include "priv-io.h"
 #include "util.h"
 
 
@@ -89,10 +89,7 @@ _gpgme_wait_on_condition (gpgme_ctx_t ctx, volatile int *cond)
 	  unsigned int idx;
 
 	  err = gpg_error_from_errno (errno);
-	  for (idx = 0; idx < ctx->fdt.size; idx++)
-	    if (ctx->fdt.fds[idx].fd != -1)
-	      _gpgme_io_close (ctx->fdt.fds[idx].fd);
-	  _gpgme_engine_io_event (ctx->engine, GPGME_EVENT_DONE, &err);
+          _gpgme_cancel_with_err (ctx, err);
 
 	  return err;
 	}
@@ -101,25 +98,23 @@ _gpgme_wait_on_condition (gpgme_ctx_t ctx, volatile int *cond)
 	{
 	  if (ctx->fdt.fds[i].fd != -1 && ctx->fdt.fds[i].signaled)
 	    {
-	      struct wait_item_s *item;
-	      
 	      ctx->fdt.fds[i].signaled = 0;
 	      assert (nr);
 	      nr--;
-	      
-	      item = (struct wait_item_s *) ctx->fdt.fds[i].opaque;
 
-	      err = item->handler (item->handler_value, ctx->fdt.fds[i].fd);
+	      LOCK (ctx->lock);
+	      if (ctx->canceled)
+		err = gpg_error (GPG_ERR_CANCELED);
+	      UNLOCK (ctx->lock);
+	      
+	      if (!err)
+		err = _gpgme_run_io_cb (&ctx->fdt.fds[i], 0);
 	      if (err)
 		{
 		  /* An error occured.  Close all fds in this context,
 		     and signal it.  */
-		  unsigned int idx;
-		  
-		  for (idx = 0; idx < ctx->fdt.size; idx++)
-		    if (ctx->fdt.fds[idx].fd != -1)
-		      _gpgme_io_close (ctx->fdt.fds[idx].fd);
-		  _gpgme_engine_io_event (ctx->engine, GPGME_EVENT_DONE, &err);
+		  _gpgme_cancel_with_err (ctx, err);
+
 		  return err;
 		}
 	    }
